import { Board } from './board/board.js'
import { Move } from './board/move.js'
import { ListView, AutoPlay } from './listview.js'
import { Chart } from './chart/stones.js'
import { Loaders } from './loaders/loaders.js'
import { FileDrop } from './filedrop.js'
import { CanvasRenderer } from './renderers/canvas.js'
import { TableRenderer } from './renderers/table.js'
import { GridRenderer } from './renderers/grid.js'
import { SvgRenderer } from './renderers/svg.js'
import { htmlEncode } from './xss.js'
import { ImportException, ParseException } from './exceptions.js'
import { Plugins } from './loaders/plugins.js'
import { LiveView } from './loaders/liveview.js'

// TODO limit this file (and others) to like ~50 LOC max

const [WIDTH, HEIGHT] = [7, 7]

class HttpException {
    constructor(response) {
        this.response = response
    }

    toString() {
        return `${ this.response.status } ${ this.response.statusText }. See console for details.`
    }
}

class FetchException {
    constructor(e, url) {
        this.e = e
        this.url = url
    }

    toString() {
        return `${ this.e.message } ${ this.url }. See console for details.`
    }
}


// plugins (enabled through data-elements)
//   - auto play (catches keyboard toggle 'a', renders UI somewhere, renderMove())
//   - move list (renderMove())
//   - live view (renderMove()), closes seconds after games ends, shows running game list and pick one that is not playing?)
//     (do not close last one? 'waiting for game...')
//   - i guess for all loaders it is true that you can pause them (which could make auto play redundant or o)



class Viewer {

    constructor($figure) {
        this.$figure = $figure
        this.timerId = null
    }

    render($figure, game, renderer, listView) {
        const $tbody = $figure.querySelector('.moves tbody')

        const renderMove = step => {
            const { move, source, destination, board } = game[step]

            renderer.draw(board)

            if (destination) {
                renderer.animateMove(new Move(source, destination))
            }
        }

        // render board on move select
        // TODO to ListView.js, onchange callback
        $tbody.addEventListener('click', e => {
            e.preventDefault()

            // TODO bug: tbody click -> fails
            let element = e.target
            while (element && element.tagName && element.tagName.toLowerCase() != 'tr') {
                element = element.parentNode
            }

            // TODO let ListView determine selectedIndex when clicked
            const step = +element.getAttribute('data-step') - 1
            ListView.selectedIndex = step
            renderMove(ListView.selectedIndex)
        })

        // auto play
        ListView.selectedIndex = 0
        const tick = () => {
            const autoplay = $figure.querySelector('input[type="checkbox"]')
            if (autoplay && autoplay.checked) {
                ListView.selectedIndex = (ListView.selectedIndex + 1) % $tbody.rows.length
                renderMove(ListView.selectedIndex)
            }
        }

        renderMove(ListView.selectedIndex)

        this.timerId = setInterval(tick, 1500)
    }
}

const load = async url => {
    try {
        const response = await fetch(url)
        if (response && !response.ok) {
            throw new HttpException(response)
        }
        return await response.text()
    } catch (e) {
        throw new FetchException(e, url)
    }
}

const attachViewer = async $figure => {
    // 1. live viewer (url= and p1= and p2 are filled)
    // 2. then default game (warn that this happened!)
    const params = new URLSearchParams(document.location.search)
    const url = $figure.getAttribute('data-url') || params.get('url') || 'games/demo.fen'
    const autoreplay = params['autoreplay']

    const plugins = new Plugins()
    const plugin = plugins.get(url)
    const liveViewer = new LiveView(null, url)
    const error = await liveViewer.open()
// TODO should this work (i.e. await on error?), maybe const promise = new Promise(resolve => resolve('resolved'), const result = await promise or use yield/next?)
//  $figure.querySelector('.flash').textContext = `${ liveViewer.constructor.name } crashed: ${ error }`

    // url or data attribute
    const view = async (moves, players, game) => {
        const viewer = new Viewer()
        const $board = $figure.querySelector('.board')

        // board
        const renderer = new TableRenderer($board.rows)
        new AutoPlay()

        // player names
        const playerNames = players.map(p => `<strong>${htmlEncode(p)}</strong>`)
        $figure.querySelector('.board caption').innerHTML =
            playerNames.join(' vs. ') || '<strong>White</strong> vs. <strong>Black<strong>'

        // move table
        const $table = $figure.querySelector('.moves')
        const $tbody = $figure.querySelector('.moves tbody')
        const headers = ['', '', 'W', 'B']
        const alignments = ['', '', 'text-right', 'text-right']
        const listView = new ListView($table, $tbody, headers, moves, alignments)

        // stone count (TODO move to move logic or after importing pgn/fen)
        const stonesWhite = [], stonesBlack = []
        for (let step = 0; step < game.length; step++) {
            const board = new Board()
            const whiteCount = board.stoneCount(game[step].board, 'o')
            const blackCount = board.stoneCount(game[step].board, 'x')

            // table
            $tbody.rows[step].cells[2].textContent = whiteCount
            $tbody.rows[step].cells[3].textContent = blackCount

            // chart
            stonesWhite.push(whiteCount)
            stonesBlack.push(blackCount)
        }

        // chart
        const $canvas = $figure.querySelector('.chart')
        const chart = new Chart($canvas)
        chart.render({
            xAxis: {
                title: 'ply'        // 0..n (~200 max?)
            },
            yAxis: {
                title: 'stones'     // 0..49
            },
            reference: {
                x: 24.5             // winning hand
            },
            legend: {
                // TODO black begins, so this is possibly playerName[1], detect from FEN who is what
                white: playerNames[0],
                black: playerNames[1]
            },
            series: [
                { name: 'black', data: stonesBlack },
                { name: 'white', data: stonesWhite }
            ]
        })

        // original game data
        $figure.querySelector('.data').textContent = text

        // render loop
        viewer.render($figure, game, renderer, listView)
    }

    const text = await load(url)

    const loader = Loaders.getLoader(url)
    const { moves, players, game } = loader.parse(text)
    view(moves, players, game)

    // file drop
    new FileDrop($figure, (filename, text) => {
        clearInterval(this.timerId)

        const loader = Loaders.getLoader(filename)
        const { moves, players, game } = loader.parse(text)
        view(moves, players, game)

        render($figure, game, renderer)
    })

    // dialogs (TODO make them work)
    const $button = $figure.querySelector('.dialog')
    $button.addEventListener('click', e => {
        const $dialog = $figure.querySelector('.about')
        $dialog.classList.toggle('hide')
    })
    const $dialog = $figure.querySelector('.about')
    $dialog.addEventListener('click', e => {
        $dialog.classList.toggle('hide')
    })
}

const loadTemplate = async selector => {
    // load template
    const response = await fetch('./templates/main.html')
    if (!response.ok) {
        throw new HttpException(response)
    }

    const template = await response.text()
    const $figures = document.querySelectorAll(selector)
    for (const $figure of $figures) {
        $figure.innerHTML = template
    }

    return $figures
}

(async () => {

    // TODO understands(url), startsWith(/) -> load .fen or .pgn (plugins or default?)



    // add stylesheet (once)
    // TODO warn if this script is included more than once, second instance -> quit
    const [href, rel] = ['./css/ataxx.css', 'stylesheet']
    const cssSelector = `link[href="${href}"][rel="${rel}"]`
    const exists = document.querySelector(cssSelector)
    if (!exists) {
        const $link = document.createElement('link')
        $link.setAttribute('href', href)
        $link.setAttribute('rel', rel)
        document.head.appendChild($link)
    }

    // viewers
    const $figures = await loadTemplate('.ataxx')       // TODO live on .ataxx (so new instances can be created and destroyed easily by just adding an element to the dom with approriate data-url)
    const style = 'padding: 2rem; text-align: center; background: #ccc; color: #4B4B4B; width: 320px'
    for (const $figure of $figures) {
        try {
            // const url = $figure.dataset('url')
            await attachViewer($figure)
        
            // TODO find the viewer instance that caused the rejection
            addEventListener('unhandledrejection', event => {
                $figure.innerHTML = `<div style="${style}">
            <strong>Ataxx viewer</strong><br>
            <p>There was trouble loading a game from <em>$ { url }</em></p>
            <p><small>Error message: ${htmlEncode(event.promise)} ${htmlEncode(event.reason)}</small></p>
            </div>`
            })

        } catch (e) {           // TODO catch all, but make sure all exceptions have a toString()
            if (e instanceof ImportException) {
                console.error(e)
            } else if (e instanceof ParseException) {
                console.error(e)
            } else if (e instanceof HttpException) {
                console.error(e)

                // TODO url undefined
                $figure.innerHTML = `<div style="${style}">
<strong>Ataxx viewer</strong><br>
<p>There was trouble loading a game from <em>${ htmlEncode(url)}</em></p>
<p><small>Error message: ${ htmlEncode(e)}</small></p>
</div>`

            } else {
                console.error(e)
                const $flash = $figure.querySelector('.flash')
                $flash.textContent = e
            }
        }
    }
})()
